﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace TestProject
{
    /// <summary>
    /// 비트맵 헬퍼
    /// </summary>
    public static class BitmapHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 비트맵 로드하기 - LoadBitmap(filePath)

        /// <summary>
        /// 비트맵 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>비트맵</returns>
        public static Bitmap LoadBitmap(string filePath)
        {
            using(Bitmap bitmap = new Bitmap(filePath))
            {
                return new Bitmap(bitmap);
            }
        }

        #endregion
        #region 소벨 3X3 필터 적용하기 - ApplySobel3X3Filter(sourceBitmap, grayscale)

        /// <summary>
        /// 소벨 3X3 필터 적용하기
        /// </summary>
        /// <param name="sourceBitmap">소스 비트맵</param>
        /// <param name="grayscale">회색조 여부</param>
        /// <returns>비트맵</returns>
        public static Bitmap ApplySobel3X3Filter(Bitmap sourceBitmap, bool grayscale = true)
        {
            Bitmap targetBitmap = ApplyConvolutionFilter
            (
                sourceBitmap,
                Matrix.Sobel3X3Horizontal,
                Matrix.Sobel3X3Vertical,
                1.0,
                0,
                grayscale
            );

            return targetBitmap;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 회선 필터 적용하기 - ApplyConvolutionFilter(sourceBitmap, xFilterArray, yFilterArray, factor, bias, grayscale)

        /// <summary>
        /// 회선 필터 적용하기
        /// </summary>
        /// <param name="sourceBitmap">소스 비트맵</param>
        /// <param name="xFilterArray">X 필터 배열</param>
        /// <param name="yFilterArray">Y 필터 배열</param>
        /// <param name="factor">팩터</param>
        /// <param name="bias">바이어스</param>
        /// <param name="grayscale">회색조 여부</param>
        /// <returns>비트맵</returns>
        private static Bitmap ApplyConvolutionFilter
        (
            Bitmap    sourceBitmap,
            double[,] xFilterArray,
            double[,] yFilterArray,
            double    factor    = 1,
            int       bias      = 0,
            bool      grayscale = false
        )
        {
            BitmapData sourceBitmapData = sourceBitmap.LockBits
            (
                new Rectangle(0, 0, sourceBitmap.Width, sourceBitmap.Height),
                ImageLockMode.ReadOnly,
                PixelFormat.Format32bppArgb
            );

            byte[] sourceByteArray = new byte[sourceBitmapData.Stride * sourceBitmapData.Height];
            byte[] targetByteArray = new byte[sourceBitmapData.Stride * sourceBitmapData.Height];

            Marshal.Copy(sourceBitmapData.Scan0, sourceByteArray, 0, sourceByteArray.Length);

            sourceBitmap.UnlockBits(sourceBitmapData);

            if(grayscale == true)
            {
                float rgb = 0;

                for(int k = 0; k < sourceByteArray.Length; k += 4)
                {
                    rgb =  sourceByteArray[k    ] * 0.11f;
                    rgb += sourceByteArray[k + 1] * 0.59f;
                    rgb += sourceByteArray[k + 2] * 0.3f;

                    sourceByteArray[k    ] = (byte)rgb;
                    sourceByteArray[k + 1] = sourceByteArray[k];
                    sourceByteArray[k + 2] = sourceByteArray[k];
                    sourceByteArray[k + 3] = 255;
                }
            }

            double blueX  = 0.0;
            double greenX = 0.0;
            double redX   = 0.0;

            double blueY  = 0.0;
            double greenY = 0.0;
            double redY   = 0.0;

            double blueTotal  = 0.0;
            double greenTotal = 0.0;
            double redTotal   = 0.0;

            int filterOffset = 1;
            int sourceOffset = 0;
            int targetOffset = 0;

            for(int offsetY = filterOffset; offsetY < sourceBitmap.Height - filterOffset; offsetY++)
            {
                for(int offsetX = filterOffset; offsetX < sourceBitmap.Width - filterOffset; offsetX++)
                {
                    blueX = greenX = redX = 0;
                    blueY = greenY = redY = 0;

                    blueTotal = greenTotal = redTotal = 0.0;

                    targetOffset = offsetY * sourceBitmapData.Stride + offsetX * 4;

                    for(int filterY = -filterOffset; filterY <= filterOffset; filterY++)
                    {
                        for(int filterX = -filterOffset; filterX <= filterOffset; filterX++)
                        {
                            sourceOffset = targetOffset + (filterX * 4) + (filterY * sourceBitmapData.Stride);

                            blueX  += (double)sourceByteArray[sourceOffset    ] *
                                      xFilterArray[filterY + filterOffset, filterX + filterOffset];

                            greenX += (double)sourceByteArray[sourceOffset + 1] *
                                      xFilterArray[filterY + filterOffset, filterX + filterOffset];

                            redX   += (double)sourceByteArray[sourceOffset + 2] *
                                      xFilterArray[filterY + filterOffset, filterX + filterOffset];

                            blueY  += (double)sourceByteArray[sourceOffset    ] *
                                      yFilterArray[filterY + filterOffset, filterX + filterOffset];

                            greenY += (double)sourceByteArray[sourceOffset + 1] *
                                      yFilterArray[filterY + filterOffset, filterX + filterOffset];

                            redY   += (double)sourceByteArray[sourceOffset + 2] *
                                      yFilterArray[filterY + filterOffset, filterX + filterOffset];
                        }
                    }

                    blueTotal  = Math.Sqrt((blueX  * blueX ) + (blueY  * blueY ));
                    greenTotal = Math.Sqrt((greenX * greenX) + (greenY * greenY));
                    redTotal   = Math.Sqrt((redX   * redX  ) + (redY   * redY  ));

                    if(blueTotal > 255)
                    {
                        blueTotal = 255;
                    }
                    else if(blueTotal < 0)
                    {
                        blueTotal = 0;
                    }

                    if(greenTotal > 255)
                    {
                        greenTotal = 255;
                    }
                    else if(greenTotal < 0)
                    {
                        greenTotal = 0;
                    }

                    if(redTotal > 255)
                    {
                        redTotal = 255;
                    }
                    else if(redTotal < 0)
                    {
                        redTotal = 0;
                    }

                    targetByteArray[targetOffset    ] = (byte)blueTotal;
                    targetByteArray[targetOffset + 1] = (byte)greenTotal;
                    targetByteArray[targetOffset + 2] = (byte)redTotal;
                    targetByteArray[targetOffset + 3] = 255;
                }
            }

            Bitmap targetBitmap = new Bitmap(sourceBitmap.Width, sourceBitmap.Height);

            BitmapData targetBitmapData = targetBitmap.LockBits
            (
                new Rectangle(0, 0, targetBitmap.Width, targetBitmap.Height),
                ImageLockMode.WriteOnly,
                PixelFormat.Format32bppArgb
            );

            Marshal.Copy(targetByteArray, 0, targetBitmapData.Scan0, targetByteArray.Length);

            targetBitmap.UnlockBits(targetBitmapData);

            return targetBitmap;
        }

        #endregion
    }
}